use strict;
use warnings;
use lib do { __FILE__ =~ m|^(.*?)[^/]*$|; "$1"; };

use Cwd;

sub init_params {

    # Purpose: set the input parameters for auto3dem. Defaults are
    # assigned for some parameters, but these can be overwritten using
    # data in the input file.
    #
    # In order to make it easier to keep track of variables,
    # parameters are collected into hashes - one describing the
    # overall flow of auto3dem and one for each of the main parallel
    # applications.
    # 
    # The "program" records of the input file have the format
    #
    #        ident  key  value
    #
    # Where the ident = (auto|all|po2r|p3dr|pcut|psf|ppft|pctfr)
    # "auto" specifies control parameters that direct the overall flow
    # of auto3dem, while "all" indicates that a parameter applies
    # to all of the parallel programs
    #
    # The "data" records of the input file have the format
    #
    #       data  datafile
    #
    # Input arguments:
    #   $param_file - input file
    #   $auto_ref   - reference to hash of control parameters
    #   $po2r_ref   - reference to hash of PO2R input parameters
    #   $p3dr_ref   - reference to hash of P3DR input parameters
    #   $pcut_ref   - reference to hash of PCUT input parameters
    #   $psf_ref    - reference to hash of PSF input parameters
    #   $ppft_ref   - reference to hash of PSF input parameters
    #   $pctfr_ref  - reference to hash of PCTFR input parameters
    #   $data_ref   - reference to array of data files
    #
    # Return values:
    #   None

    my ($param_file, $auto_ref, $po2r_ref, $p3dr_ref,
	$pcut_ref, $psf_ref, $ppft_ref, $pctfr_ref, $data_ref) = @_;

    open(IN, $param_file) || die
	"Unable to open file $param_file";

    # Set default control parameters
    $$auto_ref{fsc_hithresh} = 0.5; # High FSC threshold
    $$auto_ref{fsc_lothresh} = 0.3; # Low FSC threshold
    $$auto_ref{res_adj}   = 0.01;   # Tweak for resolution adjustment
    $$auto_ref{gauss_adj} = 0.01;   # Tweak for gaussian adjustment
    $$auto_ref{have_map}  = 1;      # Have starting map
    $$auto_ref{new_ptles} = 0;      # Don't refine ptles rel to existing map
    $$auto_ref{freeze_annulus} = 0; # Don't freeze radii of annulus
    $$auto_ref{freeze_res}     = 0; # Don't freeze resolutions
    $$auto_ref{delete_maps}    = 1; # Delete intermediate maps
    $$auto_ref{restart} = 0;        # Not a restart run
    $$auto_ref{iter_start} = 1;     # Normally start at iteration 1
    $$auto_ref{term_search} = 0;    # Don't terminate search mode runs   
    $$auto_ref{term_refine} = 0;    # Don't terminate refine mode runs
    $$auto_ref{switch_mode} = 1 ;   # Allow switch from search to refine
    $$auto_ref{bin_reduce} = 0;     # Auto reduction of bin factor
    $$auto_ref{refine_ctf}  = 1;    # Refine CTF
    $$auto_ref{quit_early} = 0;     # Quit early if PSF res_max reached
    $$auto_ref{rundir} = 'dat';     # Look for data in dat
    $$auto_ref{noctf} = 0;          # Allow CTF corrections to be applied

    $$auto_ref{estimate_res} =  1;          # Do resolution estimation
    $$auto_ref{theta_reject_lower} =  360;  # Accepting all orientations
    $$auto_ref{theta_reject_upper} = -360;  # Accepting all orientations
    $$auto_ref{phi_reject_lower}   =  360;  # Accepting all orientations
    $$auto_ref{phi_reject_upper}   = -360;  # Accepting all orientations
    $$auto_ref{box_center_offset}  = 10000; # Accepting all origins
    $$auto_ref{nselect_offset}     = 0;     # Single selection criterion
    $$auto_ref{select_delta}       = 0;     # Zero perturbation of criteria
    $$auto_ref{global_select}      = 1;     # Get select crit from all files

    $$auto_ref{hollow_map}         = 0;     # Don't hollow map
    $$auto_ref{hollow_cut_step}    = 12;    # Number of steps used in masking
    $$auto_ref{hollow_cut_weight}  = 0.001; # Weight used in masking

    # Set default program parameters
    $$psf_ref{res_step} = 50;              # Number of steps for FSC curve
    $$psf_ref{res_min} = 60;               # Lower res limit for FSC curve

    $$po2r_ref{symm_code}     = 532;       # Icosahedral symmetry
    $$po2r_ref{dcenter}       = 0.1;       # Step size for origin refinement
    $$po2r_ref{nangle}        = 4;         # Number of angle steps in each dir
    $$po2r_ref{ncenter}       = 4;         # Number of origin steps in each dir
    $$po2r_ref{tempfac}       = 0;         # Temperature factor
    $$po2r_ref{zero_fill}     = 1;         # Zero fill
    $$po2r_ref{ctf_ff1}       = 0.5;       # CTF factor 1
    $$po2r_ref{ctf_ff2}       = 0.1;       # CTF factor 2
    $$po2r_ref{ctfmode}       = 1;         # CTF mode
    $$po2r_ref{filter}        = 1;         # Filter mode
    $$po2r_ref{funcmode}      = 1;         # Funcmode
    $$po2r_ref{funcweight}    = 0;         # Funcweight
    $$po2r_ref{view_perread}  = 10;        # Views per read

    $$pctfr_ref{dangle}        = 0.05;     # Step size for CTF refinement
    $$pctfr_ref{nangle}        = 4;        # Number of steps in each dir
    $$pctfr_ref{tempfac}       = 0;        # Temperature factor
    $$pctfr_ref{zero_fill}     = 1;        # Zero fill
    $$pctfr_ref{ctf_ff1}       = 0.5;      # CTF factor 1
    $$pctfr_ref{ctf_ff2}       = 0.1;      # CTF factor 2
    $$pctfr_ref{ctfmode}       = 1;        # CTF mode
    $$pctfr_ref{filter}        = 1;        # Filter mode
    $$pctfr_ref{funcmode}      = 3;        # Funcmode (phase method)
    $$pctfr_ref{funcweight}    = 0;        # Funcweight

    $$p3dr_ref{symm_code}    = 532;        # Icosahedral symmetry
    $$p3dr_ref{tempfac}      = 0;          # temperature factor
    $$p3dr_ref{zero_fill}    = 1;          # Zero fill
    $$p3dr_ref{ctf_ff1}      = 0.05;       # 1st CTF factor
    $$p3dr_ref{ctf_ff2}      = 0.1;        # 2nd CTF factor
    $$p3dr_ref{ctfmode}      = 1;          # CTF mode
    $$p3dr_ref{filter}       = 1;          # Filter mode
    $$p3dr_ref{view_perread} = 20;         # Views per read

    $$ppft_ref{symmetry}     = 532;        # Icosahedral symmetry
    $$ppft_ref{bin_factor}   = 1;          # No binning
    $$ppft_ref{prj_filename} = "pft.prjs"; # Projections file name
    $$ppft_ref{pft_filename} = "pft.pfts"; # Polar FT file name

    $$ppft_ref{ctfmode}            = 3;    # CTF mode (divide images by CTF)
    $$ppft_ref{delta_theta}        = 0.5;  # Angular step size
    $$ppft_ref{filter_factor_1}    = 0.1;  # CTF filter factor
    $$ppft_ref{input_mode}         = 2;    # Input mode
    $$ppft_ref{jcut}               = 1;    # Minimum order Bessel function
    $$ppft_ref{mag_cen}            = 1.0;  # Midpoint for mag scale search
    $$ppft_ref{mag_norm}           = 1;    # Flag for normalizing mag
    $$ppft_ref{mag_num}            = 1;    # Extent of mag scale search
    $$ppft_ref{mag_step}           = 0.0;  # Grid sise of mag scale search
    $$ppft_ref{pftrad_step}        = 1.0;  # Radial step size in CC calculation
    $$ppft_ref{pftrad_lo}          = 1.0;  # Lower radii for CC calculation
    $$ppft_ref{sigcut}             = 0.0;  # Threshold for variance mask
    $$ppft_ref{temperature_factor} = 0.0;  # Temperature factor
    $$ppft_ref{verbose}            = 2;    # Verbosity

    $$pcut_ref{cut_step} = 12;             # Number of steps used in masking
    $$pcut_ref{cut_weight} = 0.001;        # Weight used in masking

    # Set default executable names
    $$pctfr_ref{bin} = "PCTFR";
    $$p3dr_ref{bin}  = "P3DR";
    $$po2r_ref{bin}  = "POR";
    $$pcut_ref{bin}  = "PCUT";
    $$psf_ref{bin}   = "PSF";
    $$ppft_ref{bin}  = "PPFT";

    # Set defaults for max number of CPUs when running programs
    $$p3dr_ref{max_cpu}  = 256;
    $$pctfr_ref{max_cpu} = 256;
    $$po2r_ref{max_cpu}  = 256;
    $$ppft_ref{max_cpu}  = 256;
    $$pcut_ref{max_cpu}  = 8;
    $$psf_ref{max_cpu}   = 8;

    # Parse input file and set input parameters
    while (<IN>) {
	# Get rid of comments ("#" to end of line)
	s/#.*//;

	# Get rid of leading and trailing whitespace
	s/^\s+//;
	s/\s+$//;

	# Skip empty lines
	unless(/\w/) {
	    next;
	}

	#Parse line on whitespace
	my $value;
	my ($ident, $key, @remainder) = split(/\s+/, $_);
	if($ident eq 'auto' && $key eq 'recipient') {
	    $value = join(' ', @remainder);
	} else {
	    $value = $remainder[0];
	}

	# Convert all identifiers to lower case; convert keys to
	# lowercase for lines that don't specify data files
	$ident =~ tr/A-Z/a-z/;
	unless($ident eq 'data') {
	    $key  =~ tr/A-Z/a-z/;
	}

	# Compatibility feature for older versions of auto3dem. 

	# Input for PO2R can be specified using 'por' or 'po2r'
	if($ident eq 'por') {
	    $ident = 'po2r';
	}
	# fsc_lothresh can also be specified as fsc_locut
	if($key eq 'fsc_locut') {
	    $key = 'fsc_lothresh';
	}
	# fsc_hithresh can also be specified as fsc_hicut
	if($key eq 'fsc_hicut') {
	    $key = 'fsc_hithresh';
	}

	# ctfmode and ctf_mode both accepted for po2r, p3dr, pctfr,
	# and ppft. (ctf_mode changed to ctfmode)
	if($key eq 'ctf_mode') {
	    $key = 'ctfmode';
	}

	# Skip lines that don't have required number of fields, that
	# have invalid keys, or that don't correspond to valid choices
	# of data or program
	if($ident eq 'auto'  ||
	   $ident eq 'po2r'  ||
	   $ident eq 'pctfr' ||
	   $ident eq 'p3dr'  ||
	   $ident eq 'pcut'  ||
	   $ident eq 'psf'   ||
	   $ident eq 'ppft'  ||
	   $ident eq 'all') {
	    unless(defined $key) {
		print "Warning parameter file line $. - ";
		print "missing key for '$ident'\n";
		next;
	    }
	    unless(defined $value) {
		print "Warning parameter file line $. - ";
		print "missing value for '$ident $key'\n";
		next;
	    }	    
	    if(deprecated_key($ident, $key)) {
		print "Note parameter file line $. - ";
		print "'$key' no longer used and is ignored for '$ident'\n";
		next;
	    }
	    unless(valid_key($ident, $key)) {
		print "Warning parameter file line $. - ";
		print "'$key' not valid key for '$ident'\n";
		next;
	    }
	} elsif ($ident eq 'data') {
	    unless(defined $key) {
		print "Warning parameter file line $. - ";
		print "missing file on 'data' line\n";
		next;
	    }
	} else {
	    print "Warning parameter file line $. - ";
	    print "'$ident' not program valid choice\n";
	    next;
	}

	if($ident eq 'all') {
	    $$po2r_ref{$key}      = $value;
	    $$pctfr_ref{$key}     = $value;
	    $$p3dr_ref{$key}      = $value;
	    $$pcut_ref{$key}      = $value;
	    $$psf_ref{$key}       = $value;
	    $$ppft_ref{$key}      = $value;
	} elsif ($ident eq 'po2r')  {
	    $$po2r_ref{$key}       = $value;
	} elsif ($ident eq 'pctfr') {
	    $$pctfr_ref{$key}      = $value;
	} elsif ($ident eq 'p3dr')  {
	    $$p3dr_ref{$key}      = $value;
	} elsif ($ident eq 'pcut')  {
	    $$pcut_ref{$key}      = $value;
	} elsif ($ident eq 'psf')  {
	    $$psf_ref{$key}       = $value;
	} elsif ($ident eq 'ppft') {
	    $$ppft_ref{$key}      = $value;
	} elsif ($ident eq 'auto') {
	    $$auto_ref{$key}      = $value;
	} elsif ($ident eq 'data') {
	    push(@{$data_ref}, $key);
	}
    }
    close(IN);

    # Tests to make sure that input data is complete and consistent
    my $err = 0;

    unless(defined $$auto_ref{mode} && 
	   ($$auto_ref{mode} eq 'search' || $$auto_ref{mode} eq 'refine')){
	print "Valid mode (search or refine) needs to be specified\n";
	$err = 1;
    }

    if($$auto_ref{new_ptles}) {
	unless(defined $$auto_ref{start_map}) {
	    print "Starting map not specified\n";
	    print "  If 'new_ptles' flag is set, need to specify map\n";
	    $err = 1;
	}
    }

    if($$auto_ref{have_map}) {
	unless(defined $$auto_ref{start_map}) {
	    print "Starting map not specified\n";
	    print "  If 'have_map' flag is set, need to specify map\n";
	    $err = 1;
	}
    }

    if($$auto_ref{hollow_map}) {
	unless(defined $$auto_ref{hollow_in_rad}) {
	    print "Need to set hollow_in_rad\n";
	    $err = 1;
	}
	unless(defined $$auto_ref{hollow_out_rad}) {
	    print "Need to set hollow_out_rad\n";
	    $err = 1;
	}
	if (defined $$auto_ref{hollow_in_rad} &&
	    defined $$auto_ref{hollow_out_rad} &&
	    $$auto_ref{hollow_in_rad} >= $$auto_ref{hollow_out_rad}) {
	    print "hollow_out_rad ($$auto_ref{hollow_out_rad}) ";
	    print "must be greater than ";
	    print "hollow_in_rad ($$auto_ref{hollow_in_rad})\n";
	    $err = 1;
	}
    }	    

    unless(defined $$p3dr_ref{res_min}) {
	print "Must specify res_min for p3dr\n";
	$err = 1;
    }

    unless(@{$data_ref} > 0) {
	print "Must specify at least one data file\n";
	$err = 1;
    }

    unless(defined $$auto_ref{niter} && $$auto_ref{niter} >= 0) {
	print "Number of iterations (auto niter) must be at least zero\n";
	$err = 1;
    }

    if($err) {
	die "Errors found in auto3dem input file\n";
    }

    # If auto noctf is set, set ctfmode to zero for all programs
    if($$auto_ref{noctf} == 1) {
	$$p3dr_ref{ctfmode}  = 0;
	$$po2r_ref{ctfmode}  = 0;
	$$pctfr_ref{ctfmode} = 0;
	$$ppft_ref{ctfmode}  = 0;
    }

    # If not specified, set outfile to name of current directory
    unless(defined $$auto_ref{outfile}) {
	$$auto_ref{outfile} = getcwd();
	$$auto_ref{outfile} =~ s/.*\///;
    }

    # Calculate resolution of P3DR at upper limit of Gaussian falloff
    unless(defined $$p3dr_ref{res_max}) {
	$$p3dr_ref{res_max} = 1/(1/$$p3dr_ref{res_min}+$$auto_ref{gauss_adj});
	$$p3dr_ref{res_max} = sprintf("%.2f", $$p3dr_ref{res_max});
    }

    # Set the upper resolution limit for the FSC curve
    unless(defined $$psf_ref{res_max}) {
	$$psf_ref{res_max} = $$p3dr_ref{res_min};
    }

    # Set the upper resolution limit for image/map comparison in PO2R
    unless(defined $$po2r_ref{res_max}) {
	$$po2r_ref{res_max} = 1/(1/$$p3dr_ref{res_min}-$$auto_ref{res_adj});
	$$po2r_ref{res_max} = sprintf("%.2f", $$po2r_ref{res_max});
    }

    # Set the upper resolution limit for image/map comparison in PCTFR
    unless(defined $$pctfr_ref{res_max}) {
	$$pctfr_ref{res_max} = 1/(1/$$p3dr_ref{res_min}-$$auto_ref{res_adj});
	$$pctfr_ref{res_max} = sprintf("%.2f", $$pctfr_ref{res_max});
    }

    # Set the upper resolution limit for image/map comparison in PPFT
    unless(defined $$ppft_ref{resolution_high}) {
	$$ppft_ref{resolution_high} = 1/
	    (1/$$p3dr_ref{res_min}-$$auto_ref{res_adj});
	$$ppft_ref{resolution_high} = 
	    sprintf("%.2f", $$ppft_ref{resolution_high});
    }

    # Set default angular step size for PO2R to 1/2 that used for PPFT
    unless(defined $$po2r_ref{dangle}) {
	if(defined $$ppft_ref{delta_theta}) {
	    $$po2r_ref{dangle} = 0.5 * $$ppft_ref{delta_theta};
	}
    }

    # Expand wildcards in names of data files in directory
    # $$auto_ref{rundir}. All UNIX wildcards are allowed.

    my @expanded_list = ();
    foreach my $data (@{$data_ref}) {
	push(@expanded_list, glob("$$auto_ref{rundir}/$data"));
    }
    @{$data_ref} = @expanded_list;
    foreach my $data (@{$data_ref}) {
	$data =~ s/.*\///;
	$data =~ s/\s+//g;
    }
    
    # If wildcards are used, make sure that they expand to at least
    # one data file
    unless(@{$data_ref} > 0) {
	print "Wildcards in data lines fail to expand to at least one file\n";
	die "Errors found in auto3dem input file\n";
    }

}

sub valid_key{

    # Purpose: determine whether or not a key is considered to be
    # valid for a given hash
    #
    # Input parameters
    #   $ident - name of program (auto, p3dr, po2r, etc.)
    #   $key  - key being tested
    #
    # Return values
    #   1 if key is valid, 0 otherwise

    my ($ident, $key) = @_;

    my %hash;

    $hash{po2r} = {symm_code => 1, bin => 1, res_max => 1, res_min => 1,
		   zero_fill => 1, ctfmode => 1, filter => 1,
		   tempfac => 1, ctf_ff1 => 1, ctf_ff2 => 1,
		   view_perread => 1, dangle => 1, nangle => 1,
		   funcmode => 1, funcweight => 1, dcenter => 1,
		   ncenter => 1, max_cpu => 1};

    $hash{pctfr} = {bin => 1, res_max => 1, res_min => 1,
		   zero_fill => 1, ctfmode => 1, filter => 1,
		   tempfac => 1, ctf_ff1 => 1, ctf_ff2 => 1,
		   dangle => 1, nangle => 1,
		   funcmode => 1, funcweight => 1, max_cpu => 1};
    
    $hash{p3dr} = {symm_code => 1, bin => 1, res_max => 1, res_min =>
		   1, zero_fill => 1, ctfmode => 1, filter => 1,
		   tempfac => 1, ctf_ff1 => 1, ctf_ff2 => 1,
		   view_perread => 1, max_cpu => 1};
    
    $hash{pcut} = {symm_code => 1, bin => 1, in_rad => 1, out_rad =>
		   1, cut_step => 1, cut_weight => 1, max_cpu => 1};
    
    $hash{psf} = {symm_code => 1, bin => 1, res_max => 1, res_min =>
		   1, res_step => 1, max_cpu => 1};

    $hash{ppft} = {bin => 1, input_mode => 1, bin_factor => 1,
		   symmetry => 1, delta_theta => 1, ctfmode => 1,
		   verbose => 1, filter_factor_1 => 1, pftrad_lo => 1,
		   pftrad_hi => 1, pftrad_step => 1, annulus_low => 1,
		   annulus_high => 1, temperature_factor => 1,
		   resolution_low => 1, resolution_high => 1, jcut =>
		   1, sigcut => 1, mag_cen => 1, mag_step => 1,
		   mag_num => 1, mag_norm => 1, model_filename => 1,
		   prj_filename => 1, pft_filename => 1};

    $hash{auto} = {mode => 1,
		   niter => 1, 
		   have_map => 1, 
		   start_map => 1,
		   rundir => 1, 
		   outfile => 1, 
		   delete_maps => 1, 
		   freeze_annulus => 1,
		   freeze_res => 1,
		   fsc_hithresh => 1, 
		   fsc_lothresh => 1, 
		   res_adj => 1, 
		   gauss_adj => 1,
		   theta_reject_lower => 1,
		   theta_reject_upper => 1,
		   phi_reject_lower => 1,
		   phi_reject_upper => 1,
		   box_center_offset => 1,
		   global_select => 1,
		   score_fraction => 1,
		   pft_cc_fraction => 1,
		   prj_cc_fraction => 1,
		   cmp_cc_fraction => 1,
		   score_nstd => 1,
		   pft_cc_nstd => 1,
		   prj_cc_nstd => 1,
		   cmp_cc_nstd => 1,
		   nselect_offset => 1,
		   select_delta => 1,
		   new_ptles => 1,
		   hollow_map => 1,
		   hollow_in_rad => 1,
		   hollow_out_rad => 1,
		   hollow_cut_step => 1,
		   hollow_cut_weight => 1,
		   estimate_res => 1,
		   restart => 1,
		   boxrad => 1,
	           iter_start => 1,
		   term_search => 1,
		   term_refine => 1,
		   switch_mode => 1,
	           recipient => 1,
                   refine_ctf => 1,
	           quit_early => 1,
		   noctf => 1,
		   bin_reduce => 1,
	       };
    
    $hash{all}  = {symm_code    => 1}; 
    
    if(defined $hash{$ident}{$key} || $ident eq 'data') {
	return 1;
    } else {
	return 0;
    }
}


sub deprecated_key{

    # Purpose: determine whether or not a key is considered to be
    # deprecated for a given hash
    #
    # Input parameters
    #   $ident - name of program (auto, p3dr, po2r, etc.)
    #   $key  - key being tested
    #
    # Return values
    #   1 if key is valid, 0 otherwise

    my ($ident, $key) = @_;

    my %hash;

    $hash{auto} = {dryrun => 1, 
	           split_ppft => 1,
		   final_map => 1
		   };
    
    if(defined $hash{$ident}{$key} || $ident eq 'data') {
	return 1;
    } else {
	return 0;
    }
}
1;
